# ----------------------------------------------------------------------
# Curses::UI::Language::french
# Maintainer: seb@H-K.fr
# ----------------------------------------------------------------------

package Curses::UI::Language::french; 
1;
__DATA__

# ----------------------------------------------------------------------
# For Curses::UI
# ----------------------------------------------------------------------

[screen_too_small]
Votre fenêtre est trop petite pour cette application.
Veuillez l'agrandir et redémarrer l'application.
Appuyez sur <CTRL+C> pour sortir...

# ----------------------------------------------------------------------
# For Curses::UI::Calendar
# ----------------------------------------------------------------------

[months]
Janvier Février Mars Avril Mai Juin Juillet
Août Septembre Octobre Novembre Décembre

[days_short] Di Lu Ma Me Je Ve Sa

# ----------------------------------------------------------------------
# For Curses::UI::Buttonbox
# ----------------------------------------------------------------------

[button_ok]      o:OK
[button_cancel]  c:Annuler
[button_yes]     y:Oui
[button_no]      n:Non

# ----------------------------------------------------------------------
# For Curses::UI::Dialog::Error
# ----------------------------------------------------------------------

[error_title]    Message d'erreur

# ----------------------------------------------------------------------
# For Curses::UI::Dialog::FileBrowser
# ----------------------------------------------------------------------

# The filebrowser title
[file_title]     Choisissez un fichier
[file_savetitle] Choisissez un fichier dans lequel enregistrer
[file_loadtitle] Choisissez un fichier à charger

# The labels for the dialog screen.
[file_path]      Chemin :
[file_file]      Fichier :
[file_mask]      Masque :

# The size of the longest label
[file_labelsize] 9

# On directory up
[file_dirup] Dossier parent

# For asking the user if a file may be overwritten
[file_overwrite_title] Confirmation
[file_overwrite_question_pre] 
Voulez-vous vraiment écraser le fichier "
[file_overwrite_question_post] 
" ?

# Errors
[file_err_opendir_pre] Impossible d'ouvrir le dossier "
[file_err_opendir_post] "
[file_err_nofileselected] Vous n'avez pas choisi de fichier !


